;Interrupt Handlers
;______________________________________________________________________________

IDT00:
	push dword 0x00
	jmp handleInterrupt_notIRQ

IDT01:
	push dword 0x01
	jmp handleInterrupt_notIRQ

IDT02:
	push dword 0x02
	jmp handleInterrupt_notIRQ

IDT03:
	push dword 0x03
	jmp handleInterrupt_notIRQ

IDT04:
	push dword 0x04
	jmp handleInterrupt_notIRQ

IDT05:
	push dword 0x05
	jmp handleInterrupt_notIRQ

IDT06:
	push dword 0x06
	jmp handleInterrupt_notIRQ

IDT07:
	push dword 0x07
	jmp handleInterrupt_notIRQ

IDT08:
	push dword 0x08
	cmp dword [esp+8],0x00000018			;Is the return CS at [esp+8]?
	je reflectToRM						; yes, can't be an error code
	jmp handleException					; no, handle it as an exception

IDT09:
	push dword 0x09
	jmp handleInterrupt

IDT0A:
	push dword 0x0A
	cmp dword [esp+8],0x00000018			;Is the return CS at [esp+8]?
	je reflectToRM						; yes, can't be an error code
	jmp handleException					; no, handle it as an exception

IDT0B:
	push dword 0x0B
	cmp dword [esp+8],0x00000018			;Is the return CS at [esp+8]?
	je reflectToRM						; yes, can't be an error code
	jmp handleException					; no, handle it as an exception

IDT0C:
	push dword 0x0C
	cmp dword [esp+8],0x00000018			;Is the return CS at [esp+8]?
	je reflectToRM						; yes, can't be an error code
	jmp handleException					; no, handle it as an exception

IDT0D:
	push dword 0x0D					;Set interrupt number
	cmp dword [esp+8],0x00000018			;Is the return CS at [esp+8]?
	je reflectToRM						; yes, can't be an error code

	push eax
	mov eax,[esp+8]					;eax = error code
	test al,0x02						;Was it a problem with the IDT?
	je .l1							; no
	shr eax,3							;eax = real interrupt number
	mov [esp+4],eax					;Set new interrupt number
	pop eax
	jmp reflectToRM					;And reflect the real interrupt to real mode

.l1:
	pop eax
	jmp handleException					; no, handle it as an exception


IDT0E:
	push dword 0x0E
	cmp dword [esp+8],0x00000018			;Is the return CS at [esp+8]?
	je reflectToRM						; yes, can't be an error code
	jmp handleException					; no, handle it as an exception

IDT0F:
	push dword 0x0F
	jmp handleInterrupt

IDT10:
	push dword 0x10
	jmp handleInterrupt_notIRQ

IDT11:
	push dword 0x11
	cmp dword [esp+8],0x00000018			;Is the return CS at [esp+8]?
	je reflectToRM						; yes, can't be an error code
	jmp handleException					; no, handle it as an exception

IDT12:
	push dword 0x12
	jmp handleInterrupt_notIRQ

IDT13:
	push dword 0x13
	jmp handleInterrupt_notIRQ

IDT14:
	push dword 0x14
	jmp handleInterrupt_notIRQ

IDT15:
	push dword 0x15
	jmp handleInterrupt_notIRQ

IDT16:
	push dword 0x16
	jmp handleInterrupt_notIRQ

IDT17:
	push dword 0x17
	jmp handleInterrupt_notIRQ

IDT18:
	push dword 0x18
	jmp handleInterrupt_notIRQ

IDT19:
	push dword 0x19
	jmp handleInterrupt_notIRQ

IDT1A:
	push dword 0x1A
	jmp handleInterrupt_notIRQ

IDT1B:
	push dword 0x1B
	jmp handleInterrupt_notIRQ

IDT1C:
	push dword 0x1C
	jmp handleInterrupt_notIRQ

IDT1D:
	push dword 0x1D
	jmp handleInterrupt_notIRQ

IDT1E:
	push dword 0x1E
	jmp handleInterrupt_notIRQ

IDT1F:
	push dword 0x1F
	jmp handleInterrupt_notIRQ


;Handle an interrupt with unknown cause
;______________________________________________________________________________

;Input
; [esp]	Interrupt number

handleInterrupt:
	push eax
	push ecx
	mov ecx,[esp+8]			;ecx = interrupt number
	sub cl,7					;cl = potential IRQ number + 1
	mov al,0x0B				;al = "read ISR" command
	out 0x20,al				;Send "read ISR" command
	in al,0xA0				;Dummy read (just a small delay)
	in al,0x20				;al = ISR for master PIC
	shr al,cl					;carry flag = set if IRQ is in service
	pop ecx
	pop eax
	jc reflectToRM				;Reflect interrupt to real mode if corresponding IRQ is in service

	cmp dword [esp],0x0F		;Could it be "spurious IRQ 7"?
	jne handleInterrupt_notIRQ	; no
	add esp,4					; yes, assume it is ("exception 0x0F" is extremely unlikely)
	iretd


handleInterrupt_notIRQ:			;Could be exception or software interrupt
	push eax
	push ecx
	mov ecx,[esp+8]			;ecx = interrupt number
	mov eax,[esp+12]			;eax = return EIP
	cmp [cs:eax-1],cl			;Is the byte at "return EIP - 1" the interrupt number?
	jne .exception				; no, couldn't have been a software interrupt
	cmp byte [cs:eax-2],0xCD		;Is the byte at "return EIP - 2" the opcode for "int n"?
	jne .exception				; no, couldn't have been a software interrupt
	pop ecx
	pop eax
	jmp reflectToRM			;Was a software interrupt

.exception:
	pop ecx
	pop eax
	push dword [esp]			;Shift interrupt number (to make room from pretend error code)
	mov dword [esp+4],0			;Set fake error code
	jmp handleException




;Reflect an interrupt back to real mode
;______________________________________________________________________________

;Input
; [esp]	Interrupt number


reflectToRM:
	push ds
	push es
	push fs
	push gs

	lidt [NULL_IDT]				;Load "NULL IDT" to ensure NMI causes triple fault
	sub esp,14					;Make space for original CR0, retf data and iret data
	push eax

;Switch to real mode

	jmp 0x08:.here1				;Switch to 16-bit code

	bits 16

.here1:
	mov ax,0x10					;ax = 16-bit data selector
	mov ss,ax
	mov ds,ax
	mov es,ax
	mov fs,ax
	mov gs,ax

	mov eax,cr0
	mov [esp+14],eax				;Save original CR0
	and eax,0x7FFFFFFE
	mov cr0,eax
	jmp .here2
.here2:
	jmp 0x0000:.here3
.here3:
	xor ax,ax
	mov ds,ax
	mov es,ax
	mov fs,ax
	mov gs,ax
	mov ss,ax

	lidt [BIOS_IDT]				;Load "BIOS IDT"

	mov ax,[esp+46]				;eax = original eflags
	and ax,~0x0200					;Make sure interrupt enable flag is clear
	mov [esp+12],ax				;Set flags for iret
	mov word [esp+8],.iretDest		;Set IP for iret
	mov word [esp+10],0				;Set CS for iret

	mov eax,[esp+34]				;eax = interrupt number
	mov eax,[cs:eax*4]				;eax = IVT segment and offset
	mov [esp+4],eax				;Set "retf" data

	pop eax						;eax = original EAX

	retf
.iretDest:
	push eax
	pushf
	pop ax
	and ax,~0x0200					;Make sure interrupt enable flag is clear
	and word [esp+36],0x0200			;Preserve original interrupt enable flag
	or [esp+36],ax					;Update original eflags (interrupt enable flag left as is)

;Switch back to protected mode

	mov eax,[esp+4]				;eax = original cr0
	lgdt [cs:GDT]					;Load GDT
	lidt [cs:NULL_IDT]				;Load "NULL IDT" to ensure NMI causes triple fault
	mov cr0,eax					;Enable protected mode
	jmp .here4					;Flush pipeline on old CPUs
.here4:
	jmp 0x0018:.here5				;Load CS and switch to 32-bit
.here5:
	bits 32

	mov ax,0x0020
	mov ss,ax

	lidt [cs:BBW_IDT]				;Load protected mode IDT

	pop eax
	add esp,4						;Clean up stack (remove old CR0)
	pop gs
	pop fs
	pop es
	pop ds
	add esp,4						;Clean up stack (remove interrupt number)
	iret


;Handle an exception
;______________________________________________________________________________
;
;Note: This just stores everything, then jumps to code to display the stored
;      values. This allows the actual display code to be overridden (e.g. if
;      you want to use graphics mode or something)
;
;Input
; [esp]	Exception number

handleException:
	push eax
	push ds
	push es

	mov ax,0x0020
	mov ds,ax
	mov es,ax

	pop dword [tempES]
	pop dword [tempDS]
	pop dword [tempEAX]
	pop dword [tempExceptionNumber]
	pop dword [tempErrorCode]

	mov eax,[esp]
	mov [tempEIP],eax
	mov eax,[esp+4]
	mov [tempCS],eax
	mov eax,[esp+8]
	mov [tempEFLAGS],eax

	mov eax,cr2
	mov [tempEBX],ebx
	mov [tempECX],ecx
	mov [tempEDX],edx
	mov [tempESI],esi
	mov [tempEDI],edi
	mov [tempEBP],ebp
	mov [tempCR2],eax
	mov [tempESP],esp
	mov eax,ss
	mov [tempSS],eax

	jmp [exceptionHandlerAddress]
